#include "spatial.h"
#include "../../functions/functions.h"

namespace Filtering { namespace AnisoTools {

static void filter_plane(const Plane<PIXTYPE> &destination, const Plane<const PIXTYPE> &source, const Plane<const PIXTYPE> &gradient, const int *strengths, int strength)
{
   for ( int y = 0; y < destination.height(); y++ )
   {
      for ( int x = 0; x < destination.width(); x++ )
      {
         const int cg = gradient[ x + y * gradient.pitch() ];
         const int c = source[ x + y * source.pitch() ];
         int val = 0;

         val += (strengths[ gradient[ x - 1 + y * gradient.pitch() ] - cg ] * (source[ x - 1 + y * source.pitch() ] - c) + (1 << (SPRECISION - 1))) >> SPRECISION;
         val += (strengths[ gradient[ x + 1 + y * gradient.pitch() ] - cg ] * (source[ x + 1 + y * source.pitch() ] - c) + (1 << (SPRECISION - 1))) >> SPRECISION;
         val += (strengths[ gradient[ x + (y + 1) * gradient.pitch() ] - cg ] * (source[ x + (y + 1) * source.pitch() ] - c) + (1 << (SPRECISION - 1))) >> SPRECISION;
         val += (strengths[ gradient[ x + (y - 1) * gradient.pitch() ] - cg ] * (source[ x + (y - 1) * source.pitch() ] - c) + (1 << (SPRECISION - 1))) >> SPRECISION;

         destination[ x + y * destination.pitch() ] = c + ((val * strength + (2 << STPRECISION)) >> (2 + STPRECISION));
      }
   }
}

static void filter_gradient(const Plane<PIXTYPE> &gradient, const Plane<const PIXTYPE> &source, int nCenter)
{
   for ( int y = 0; y < gradient.height(); y++ )
   {
      for ( int x = 0; x < gradient.width(); x++ )
      {
         int val = 0;

         for (int i = -1; i < 2; i++ )
            for (int j = -1; j < 2; j++ )
               val += (i == 0 && j == 0 ? nCenter : 1) * source[x + i + (y + j) * source.pitch()];

         gradient[x + y * gradient.pitch()] = (val + (nCenter + 8) / 2) / (nCenter + 8);
      }
   }
}

Frame<Byte> Filters::Spatial::Filter::get_frame(int n, const Frame<Byte> &output_frame)
{
   Functions::ConvertFrame<PIXTYPE, Byte>( frames[0].select(1, 1), childs[0]->get_const_frame( n ), PRECISION );

   for ( int j = 0; j < nIterations; j++ )
   {
      Functions::PaddFrame<PIXTYPE>( frames[j & 1], 1, 1 );
      for ( int i = 0; i < plane_counts[C]; i++ )
      {
         filter_gradient( gradient.select(1, 1).plane(i), frames[j & 1].select(1, 1).plane(i), nCenter );
         Functions::PaddFrame<PIXTYPE>( gradient, 1, 1 );
         filter_plane( frames[1 - (j & 1)].select(1, 1).plane(i), frames[j & 1].select(1, 1).plane(i), gradient.select(1, 1).plane(i), pEdges[i], pSteps[i]);
      }
   }

   Functions::ConvertFrame<Byte, PIXTYPE>( output_frame, Frame<const PIXTYPE>(frames[nIterations & 1].select(1, 1)), -PRECISION );

   return output_frame;
}

} } // namespaces AnisoTools, Filtering
